#!/usr/bin/env python
import hysop
import hysop.gpu
hysop.gpu.CL_PROFILE = True
from hysop.constants import np, HDF5
from hysop.mpi.main_var import MPI
from hysop.tools.parameters import Discretization, IOParams
from hysop.methods_keys import TimeIntegrator, Interpolation, Remesh,\
    Support, Splitting, MultiScale, MultiScale, ExtraArgs
from hysop.numerics.integrators.runge_kutta2 import RK2 as RK
from hysop.numerics.remeshing import L2_1, L4_2, L2_1, L4_4, L8_4, L6_4, Linear
from hysop.operator.advection import Advection
from hysop.operator.analytic import Analytic
from hysop.operator.custom import CustomMonitor
from hysop.operator.hdf_io import HDF_Writer
from hysop.gpu.gpu_transfer import DataTransfer
from hysop.problem.problem import Problem
sin, cos, pi, sqrt = np.sin, np.cos, np.pi, np.sqrt


def vitesse(res, x, y, t=0.):
    res[0][...] = -np.sin(x * np.pi) ** 2 * np.sin(y * np.pi * 2) * \
        np.cos(t * np.pi / 3.)
    res[1][...] = np.sin(y * np.pi) ** 2 * np.sin(x * np.pi * 2) * \
        np.cos(t * np.pi / 3.)
    return res


def scalaire(res, x, y, t=0.):
    rr = np.sqrt((x - 0.5) ** 2 + (y - 0.75) ** 2)
    res[0][...] = 0.
    res[0][rr < 0.15] = 1.
    return res


def volume(_simu, var):
    return np.sum(var[0].data[0] > 0.5) * \
        np.prod(var[0].topology.mesh.space_step)


# Parameters
dim = 2
boxLength = [1., 1.]
boxMin = [0., 0.]
nbElem_v = [1025] * dim
nbElem_s = [1025] * dim
dv = Discretization(nbElem_v)
ds = Discretization(nbElem_s)

timeStep = 0.35 / (2. * np.pi)
finalTime = 12. - timeStep
outputModulo = 20
simu = hysop.Simulation(tinit=0.0, tend=finalTime,
                        timeStep=timeStep, iterMax=100000)

# Domain
box = hysop.Box(length=boxLength, origin=boxMin)

# Fields
scal = hysop.Field(domain=box, name='Scalar')  # formula=scalaire
velo = hysop.Field(domain=box, name='Velocity',
                   is_vector=True)  # formula=scalaire
# Operators
advec_method = {TimeIntegrator: RK,
                Interpolation: Linear,
                Remesh: L2_1,
                Support: 'gpu_2k',
                Splitting: 'o2',
                MultiScale: Linear}
advec_method[ExtraArgs] = {'user_src': ['./levelSet2D.cl']}
# advec_method = {Scales: 'p_64', MultiScale: 'L4_4'}

advec = Advection(velo,
                  discretization=dv,
                  variables={scal: ds},
                  method=advec_method)
advec.discretize()

# Get topologies from operator
topo_v = advec.advec_dir[0].discreteFields[velo].topology
topo_s = advec.advec_dir[0].discreteFields[scal].topology

velocity = Analytic(variables={velo: topo_v},
                    method={Support: 'gpu'})

volume_m = CustomMonitor(function=volume, res_shape=1,
                         variables={scal: topo_s},
                         io_params=IOParams(filename="volume.dat"))
p = HDF_Writer(variables={scal: topo_s},
               io_params=IOParams(frequency=outputModulo,
                                  filename="levelset",
                                  fileformat=HDF5))
# ToHost frequency is set to 1 since the volume computation
# is performed at each iteration.
toHost = DataTransfer(source=topo_s, target=p,
                      variables={scal: topo_s},
                      freq=1,
                      run_till=[p, volume_m])

pb = Problem([velocity, advec, toHost, p, volume_m],
             simu, dumpFreq=-1)

pb.setup()
scal.initialize(topo=topo_s)
velo.initialize(topo=topo_v)

scal.discreteFields[topo_s].toHost()
scal.discreteFields[topo_s].wait()
volume_m.apply(simu)
p.apply(simu)

ctime = MPI.Wtime()
pb.solve()
print MPI.Wtime() - ctime

simu.finalize()
scal.discreteFields[topo_s].toHost()
scal.discreteFields[topo_s].wait()
p.apply(simu)
volume_m.apply(simu)

pb.finalize()

print pb.profiler
