#!/usr/bin/env python
# Scripts arguments:
# 1. Flow resolution
# 2. Scalar resolution
# 3. Dictionary for devices id: (mpi rank: device id)
# 4. Is the initial condition is perturbed
# 5. Is data output
import sys
USER_NB_ELEM_UW = eval(sys.argv[1])
USER_NB_ELEM_S = eval(sys.argv[2])
USER_RANK_DEVICE_ID = eval(sys.argv[3])
RANDOM_INIT = eval(sys.argv[4])
IS_OUTPUT = eval(sys.argv[5])
import hysop
# import hysop.gpu
# hysop.gpu.CL_PROFILE = True
from hysop.constants import np, HDF5, ASCII, HYSOP_MPI_REAL
from hysop.mpi.main_var import MPI, main_size, main_rank, main_comm
from hysop.tools.parameters import MPIParams, Discretization, IOParams
from hysop.problem.simulation import Simulation
from hysop.fields.variable_parameter import VariableParameter
from hysop.methods_keys import TimeIntegrator, Interpolation, Remesh,\
    Support, Splitting, MultiScale, MultiScale, SpaceDiscretisation, \
    GhostUpdate, Scales, dtCrit, ExtraArgs
from hysop.numerics.integrators.runge_kutta2 import RK2 as RK
from hysop.numerics.integrators.runge_kutta3 import RK3 as RK3
from hysop.numerics.finite_differences import FD_C_4
from hysop.numerics.remeshing import L6_4 as remesh_formula
from hysop.numerics.remeshing import Linear
from hysop.operator.advection import Advection
from hysop.operator.diffusion import Diffusion
from hysop.operator.stretching import Stretching
from hysop.operator.poisson import Poisson
from hysop.operator.differential import Curl
from hysop.operator.adapt_timestep import AdaptTimeStep
from hysop.operator.custom import CustomMonitor
from hysop.operator.redistribute_inter import RedistributeInter
from hysop.operator.redistribute_intra import RedistributeIntra
from hysop.gpu.gpu_transfer import DataTransfer
from hysop.domain.subsets.boxes import SubBox
from hysop.operator.hdf_io import HDF_Writer
from hysop.operator.energy_enstrophy import EnergyEnstrophy
import hysop.tools.numpywrappers as npw
pi = np.pi
cos = np.cos
sin = np.sin
exp = np.exp
abs = np.abs
tanh = np.tanh


TASK_UW = 1
TASK_SCALAR = 2
PROC_TASKS = [TASK_UW, ] * main_size
for p in USER_RANK_DEVICE_ID:
    PROC_TASKS[p] = TASK_SCALAR
try:
    DEVICE_ID = USER_RANK_DEVICE_ID[main_rank]
except KeyError:
    DEVICE_ID = None
out_freq = 200
# Physical parameters:
# Flow viscosity
VISCOSITY = 1e-4
# Schmidt number
SC = ((1. * USER_NB_ELEM_S[0] - 1.)/(1. * USER_NB_ELEM_UW[0] - 1.))**2
# Scalar diffusivity
DIFF_COEFF_SCAL = VISCOSITY / SC

width = 0.01
ampl3 = 0.3
ampl = 0.05
ampl2 = 0.05


ctime = MPI.Wtime()

# Domain
box = hysop.Box(length=[1., 1., 1.], origin=[0., 0., 0.],
                proc_tasks=PROC_TASKS)
mpi_params = MPIParams(comm=box.comm_task, task_id=PROC_TASKS[main_rank])
mpi_params_S = MPIParams(comm=box.comm_task, task_id=TASK_SCALAR)
mpi_params_UW = MPIParams(comm=box.comm_task, task_id=TASK_UW)


def computeVel(res, x, y, z, t):
    yy = abs(y - 0.5)
    aux = (0.1 - 2. * yy) / (4. * width)
    strg1 = exp(-abs(aux ** 2))
    strg2 = exp(-abs(aux ** 2))
    strg3 = exp(-abs(aux ** 2))
    if RANDOM_INIT:
        from create_random_arrays import random_init
        randX, randY, randZ = random_init(res[0].shape, box.comm_task)
        strg1 = exp(-abs(aux ** 2)) * randX
        strg2 = exp(-abs(aux ** 2)) * randY
        strg3 = exp(-abs(aux ** 2)) * randZ
    else:
        strg1 = 0.
        strg2 = 0.
        strg3 = 0.
    res[0][...] = 0.5 * (1. + tanh(aux))
    res[0][...] *= (1. + ampl3 * sin(8. * pi * x))
    res[0][...] *= (1. + ampl * strg1)
    res[1][...] = ampl * strg2
    res[2][...] = ampl * strg3
    return res


def initScal(res, x, y, z, t):
    yy = abs(y - 0.5)
    aux = (0.1 - 2. * yy) / (4. * width)
    res[0][...] = 0.5 * (1. + tanh(aux))
    res[0][...] *= (1. + ampl3 * sin(8. * pi * x))
    return res


temp_maxvelo = npw.zeros((3, ))
maxvelo_values = npw.zeros((3, ))


def calc_maxvelo(simu, v):
    temp_maxvelo[0] = np.max(np.abs(v[0].data[0]))
    temp_maxvelo[1] = np.max(np.abs(v[0].data[1]))
    temp_maxvelo[2] = np.max(np.abs(v[0].data[2]))
    v[0].topology.comm.Allreduce(sendbuf=[temp_maxvelo, 3, HYSOP_MPI_REAL],
                                 recvbuf=[maxvelo_values, 3, HYSOP_MPI_REAL],
                                 op=MPI.MAX)
    return maxvelo_values


# Fields
velo = hysop.Field(domain=box, formula=computeVel,
                   name='Velocity', is_vector=True)
vorti = hysop.Field(domain=box,
                    name='Vorticity', is_vector=True)
scal = hysop.Field(domain=box, formula=initScal,
                   name='Scalar', is_vector=False)

data = {'dt': 0.001}
dt = VariableParameter(data)
simu = Simulation(tinit=0.0, tend=5., timeStep=0.001, iterMax=10000)

# Flow discretizations:
d_uw = Discretization(USER_NB_ELEM_UW)
d_uw_ghosts = Discretization(USER_NB_ELEM_UW, [2, ] * 3)
# Scalar discretization
d_s = Discretization(USER_NB_ELEM_S)
# Velocity discretization for scalar advection
if USER_NB_ELEM_UW[0] == USER_NB_ELEM_S[0]:
    d_uw_for_s = Discretization(USER_NB_ELEM_UW)
else:
    d_uw_for_s = Discretization(USER_NB_ELEM_UW, [1, ] * 3)

# Topologies
topo_GPU_scal = None
topo_GPU_velo = None
topo_CPU_velo_fft = None
topo_CPU_velo_ghosts = None
topo_CPU_velo_scales = None
if box.isOnTask(TASK_UW):
    topo_CPU_velo_scales = box.create_topology(
        d_uw, dim=2, mpi_params=mpi_params_UW)
if box.isOnTask(TASK_SCALAR):
    topo_GPU_velo = box.create_topology(
        d_uw_for_s, dim=1, mpi_params=mpi_params_S)
    topo_GPU_scal = box.create_topology(
        d_s, dim=1, mpi_params=mpi_params_S)

# Operators
# GPU operators
advec_scal_method = {TimeIntegrator: RK,
                     Interpolation: Linear,
                     Remesh: remesh_formula,
                     Support: 'gpu_1k',
                     Splitting: 'o2',
                     MultiScale: Linear}
if PROC_TASKS.count(TASK_SCALAR) > 1:
    advec_scal_method[ExtraArgs] = {'device_id': DEVICE_ID,
                                    'device_type': 'gpu'}
else:
    advec_scal_method[ExtraArgs] = {'max_velocity': [1.2, 0.7, 0.7],
                                    'max_dt': 0.012,
                                    'device_id': DEVICE_ID,
                                    'device_type': 'gpu'}
advec_scal = Advection(velo,
                       discretization=topo_GPU_velo,
                       variables={scal: topo_GPU_scal},
                       mpi_params=mpi_params_S,
                       method=advec_scal_method)
diffusion_scal = Diffusion(viscosity=DIFF_COEFF_SCAL,
                           vorticity=scal,
                           discretization=topo_GPU_scal,
                           mpi_params=mpi_params_S,
                           method={Support: 'gpu',
                                   SpaceDiscretisation: 'fd',
                                   ExtraArgs: {'device_id': DEVICE_ID,
                                               'device_type': 'gpu'}})
diffusion_scal.name += '_(Scalar)'

# CPU operators
advec = Advection(velo,
                  discretization=topo_CPU_velo_scales,
                  variables={vorti: topo_CPU_velo_scales},
                  mpi_params=mpi_params_UW,
                  method={Scales: 'p_64', MultiScale: 'L4_4'})
stretch = Stretching(velo, vorti, discretization=d_uw_ghosts,
                     mpi_params=mpi_params_UW)
diffusion = Diffusion(variables={vorti: d_uw},
                      viscosity=VISCOSITY,
                      mpi_params=mpi_params_UW)
poisson = Poisson(velo, vorti, discretization=d_uw,
                  mpi_params=mpi_params_UW)
c = Curl(velo, vorti, discretization=d_uw,
         method={SpaceDiscretisation: 'fftw', GhostUpdate: True},
         mpi_params=mpi_params_UW)
#dt_output = None
#if IS_OUTPUT:
dt_output = IOParams(frequency=1, filename='dt.dat', fileformat=ASCII)
dt_adapt = AdaptTimeStep(velo, vorti,
                         simulation=simu,
                         time_range=[0, np.infty],
                         discretization=d_uw_ghosts,
                         method={TimeIntegrator: RK3,
                                 SpaceDiscretisation: FD_C_4,
                                 dtCrit: ['gradU', 'cfl']},
                         lcfl=0.15,
                         cfl=1.5,
                         io_params=dt_output,
                         mpi_params=mpi_params_UW)

# Operators discretizations
if box.isOnTask(TASK_SCALAR):
    for op in (advec_scal, diffusion_scal):
        op.discretize()
if box.isOnTask(TASK_UW):
    for op in (advec, stretch, diffusion, poisson, c, dt_adapt):
        op.discretize()

# Get some topologies
if box.isOnTask(TASK_UW):
    topo_CPU_velo_fft = poisson.discreteFields[velo].topology
    topo_CPU_velo_ghosts = stretch.discreteFields[velo].topology

if IS_OUTPUT:
    # Defining subdomains
    L_V = 1. - 1. / (1. * USER_NB_ELEM_UW[0])
    L_S = 1. - 1. / (1. * USER_NB_ELEM_S[0])
    XY_plane_v = SubBox(origin=[0., 0., 0.5], length=[L_V, L_V, 0.],
                        parent=box)
    XZ_plane_v = SubBox(origin=[0., 0.5, 0.], length=[L_V, 0., L_V],
                        parent=box)
    XY_plane_s = SubBox(origin=[0., 0., 0.5], length=[L_S, L_S, 0.],
                        parent=box)
    XZ_plane_s = SubBox(origin=[0., 0.5, 0.], length=[L_S, 0., L_S],
                        parent=box)
    # Defining output operators
    p_velo = HDF_Writer(variables={velo: topo_CPU_velo_fft},
                        mpi_params=mpi_params_UW,
                        io_params=IOParams(frequency=out_freq,
                                           filename='flow',
                                           fileformat=HDF5))
    p_velo_xy = HDF_Writer(variables={velo: topo_CPU_velo_fft,
                                      vorti: topo_CPU_velo_fft},
                           var_names={velo: 'Velocity', vorti: 'Vorticity'},
                           subset=XY_plane_v,
                           mpi_params=mpi_params_UW,
                           io_params=IOParams(frequency=out_freq,
                                              filename='flow_XY',
                                              fileformat=HDF5))
    p_velo_xz = HDF_Writer(variables={velo: topo_CPU_velo_fft,
                                      vorti: topo_CPU_velo_fft},
                           var_names={velo: 'Velocity', vorti: 'Vorticity'},
                           subset=XZ_plane_v,
                           mpi_params=mpi_params_UW,
                           io_params=IOParams(frequency=out_freq,
                                              filename='flow_XZ',
                                              fileformat=HDF5))
    p_scal_xy = HDF_Writer(variables={scal: topo_GPU_scal},
                           var_names={scal: 'Scalar'},
                           subset=XY_plane_s,
                           mpi_params=mpi_params_S,
                           io_params=IOParams(frequency=out_freq,
                                              filename='scal_XY',
                                              fileformat=HDF5))
    p_scal_xz = HDF_Writer(variables={scal: topo_GPU_scal},
                           var_names={scal: 'Scalar'},
                           subset=XZ_plane_s,
                           mpi_params=mpi_params_S,
                           io_params=IOParams(frequency=out_freq,
                                              filename='scal_XZ',
                                              fileformat=HDF5))
    p_scal_xy.name += '_(Scalar)'
    p_scal_xz.name += '_(Scalar)'
    energy = EnergyEnstrophy(velocity=velo,
                             vorticity=vorti,
                             discretization=topo_CPU_velo_fft,
                             mpi_params=mpi_params_UW,
                             io_params=IOParams(frequency=1,
                                                filename='energy.dat',
                                                fileformat=ASCII))
    maxvelo = CustomMonitor(variables={velo: topo_CPU_velo_scales},
                            function=calc_maxvelo,
                            res_shape=3,
                            mpi_params=mpi_params_UW,
                            io_params=IOParams(frequency=1,
                                               filename='maxvelo.dat',
                                               fileformat=ASCII))
    if box.isOnTask(TASK_UW):
        for op in (p_velo, p_velo_xy, p_velo_xz, energy, maxvelo):
            op.discretize()
    if box.isOnTask(TASK_SCALAR):
        for op in (p_scal_xy, p_scal_xz):
            op.discretize()

# Redistribute operators
# CPU redistributes
toGhost_vorti = RedistributeIntra(variables=[vorti],
                                  source=advec, target=stretch,
                                  mpi_params=mpi_params_UW)
toGhost_vorti.name += '_toG_W'
toGhost_velo = RedistributeIntra(variables=[velo],
                                 source=poisson, target=stretch,
                                 run_till=[stretch, dt_adapt],
                                 mpi_params=mpi_params_UW)
toGhost_velo.name += '_toG_V'
toScales_velo = RedistributeIntra(variables=[velo],
                                  source=poisson, target=advec,
                                  mpi_params=mpi_params_UW)
toScales_velo.name += '_toScales_V'
toScales_vorti = RedistributeIntra(variables=[vorti],
                                   source=poisson, target=advec,
                                   mpi_params=mpi_params_UW)
toScales_vorti.name += '_toScales_W'
fromGhost_vorti = RedistributeIntra(variables=[vorti],
                                    source=stretch, target=diffusion,
                                    mpi_params=mpi_params_UW)
fromGhost_vorti.name += '_FromG_W'
toDev = DataTransfer(source=topo_GPU_velo, target=advec_scal,
                     variables={velo: topo_GPU_velo},
                     mpi_params=mpi_params_S)
toDev.name += '_ToDev_V'

redistrib = RedistributeInter(variables=[velo],
                              parent=main_comm,
                              source=topo_CPU_velo_fft, target=topo_GPU_velo,
                              source_id=TASK_UW, target_id=TASK_SCALAR,
                              run_till=[toDev])
if IS_OUTPUT:
    toHost = DataTransfer(source=topo_GPU_scal, target=p_scal_xy,
                          variables={scal: topo_GPU_scal},
                          mpi_params=mpi_params_S,
                          freq=out_freq,
                          run_till=[p_scal_xz, ])
    toHost.name += '_ToHost_S'

# Operators setup
if box.isOnTask(TASK_SCALAR):
    for op in (advec_scal, diffusion_scal):
        op.setup()
    if IS_OUTPUT:
        for op in (p_scal_xy, p_scal_xz, toHost):
            op.setup()
    for op in (toDev, ):
        op.setup()
if box.isOnTask(TASK_UW):
    for op in (advec, stretch, diffusion, poisson, c, dt_adapt):
        op.setup()
    if IS_OUTPUT:
        for op in (p_velo, p_velo_xy, p_velo_xz, energy, maxvelo):
            op.setup()
    for op in (toGhost_vorti, fromGhost_vorti, toGhost_velo,
               toScales_velo, toScales_vorti):
        op.setup()
# Wait for all operators setup before setup the intra-comm redistribute
main_comm.Barrier()
redistrib.setup()

# Operators list
if IS_OUTPUT:
    operators_list = [redistrib,
                      toDev, advec_scal, diffusion_scal,
                      toHost, p_scal_xy, p_scal_xz,
                      advec, toGhost_vorti, stretch,
                      fromGhost_vorti, diffusion, poisson,
                      p_velo, p_velo_xy, p_velo_xz, energy, maxvelo,
                      toGhost_velo, toScales_velo, toScales_vorti, dt_adapt]
else:
    operators_list = [redistrib,
                      toDev, advec_scal, diffusion_scal,
                      advec, toGhost_vorti, stretch,
                      fromGhost_vorti, diffusion, poisson,
                      toGhost_velo, toScales_velo, toScales_vorti, dt_adapt]

# Fields initializations
if box.isOnTask(TASK_SCALAR):
    scal.initialize(topo=topo_GPU_scal)
    advec_dirX = advec_scal.advec_dir[0].discrete_op
    gpu_scal = advec_dirX.fields_on_grid[0]
    gpu_pscal = advec_dirX.fields_on_part[gpu_scal][0]
    diffusion_scal.discrete_op.set_field_tmp(gpu_pscal)
    if IS_OUTPUT:
        p_scal_xy.apply(simu)
        p_scal_xz.apply(simu)
if box.isOnTask(TASK_UW):
    velo.initialize(topo=topo_CPU_velo_fft)
    c.apply(simu)
    poisson.apply(simu)
    toGhost_velo.apply(simu)
    toScales_velo.apply(simu)
    toScales_vorti.apply(simu)
    if IS_OUTPUT:
        p_velo.apply(simu)
        p_velo_xy.apply(simu)
        p_velo_xz.apply(simu)

simu.initialize()
setup_time = MPI.Wtime() - ctime
main_comm.Barrier()

# Solve
solve_time = 0.
while not simu.isOver:
    ctime = MPI.Wtime()
    if main_rank == 0:
        simu.printState()
    for op in operators_list:
        if box.isOnTask(op.task_id()):
            op.apply(simu)
    if box.isOnTask(TASK_SCALAR):
        # Wait gpu operations on scalar
        advec_scal.advec_dir[0].discreteFields[scal].wait()
    solve_time += MPI.Wtime() - ctime
    dt_adapt.wait()
    # Synchronize threads
    main_comm.Barrier()
    simu.advance()

main_comm.Barrier()
nb_ite = simu.currentIteration
simu.finalize()


if IS_OUTPUT:
    if box.isOnTask(TASK_SCALAR):
        p_scal_xy.apply(simu)
        p_scal_xz.apply(simu)
    if box.isOnTask(TASK_UW):
        p_velo.apply(simu)
        p_velo_xy.apply(simu)
        p_velo_xz.apply(simu)

for op in operators_list:
    if box.isOnTask(op.task_id()):
        op.finalize()

velo.finalize()
if box.isOnTask(TASK_SCALAR):
    scal.finalize()
if box.isOnTask(TASK_UW):
    vorti.finalize()
main_comm.Barrier()
